from datetime import datetime
from typing import Any, Dict, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.errors import Invalid


class Annotation(Struct, frozen=True, omit_defaults=True):
    """Annotation class.

    It is used to convey extra information to describe any
    SDMX construct.
    This information may be in the form of a URL reference and/or
    a multilingual text (represented by the association to
    InternationalString).

    Attributes:
        id: The identifier of the annotation.
        title: The title of the annotation.
        text: The text of the annotation.
        url: The URL of the annotation.
        type: The type of the annotation.
    """

    id: Optional[str] = None
    title: Optional[str] = None
    text: Optional[str] = None
    url: Optional[str] = None
    type: Optional[str] = None

    def __post_init__(self) -> None:
        """Additional validation checks for Annotation."""
        if (
            not self.id
            and not self.title
            and not self.text
            and not self.url
            and not self.type
        ):
            raise Invalid(
                "Empty annotation",
                (
                    "All fields of the annotation have been left empty."
                    "Please set at least one."
                ),
            )

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__annotations__.keys():
            v = self.__getattribute__(k)
            if v:
                out.append(f"{k}={str(v)}")
        return ", ".join(out)


class AnnotableArtefact(Struct, frozen=True, omit_defaults=True, kw_only=True):
    """Annotable Artefact class.

    Superclass of all SDMX artefacts.
    Contains the list of annotations.

    Attributes:
        annotations: The list of annotations attached to the artefact.
    """

    annotations: Sequence[Annotation] = ()

    @classmethod
    def __all_annotations(cls) -> Dict[str, Any]:
        class_attributes = {}
        for c in cls.__mro__:
            if "__annotations__" in c.__dict__:
                class_attributes.update(c.__annotations__)
        return dict(reversed(list(class_attributes.items())))

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__all_annotations().keys():
            v = self.__getattribute__(k)
            if v:
                out.append(f"{k}={str(v)}")
        return ", ".join(out)


class IdentifiableArtefact(AnnotableArtefact, frozen=True, omit_defaults=True):
    """Identifiable Artefact class.

    Provides identity to all derived classes. It also provides annotations
    to derive classes because it is a subclass of AnnotableArtefact.

    Attributes:
        id: The identifier of the artefact.
        uri: The URI of the artefact.
        urn: The URN of the artefact.
    """

    id: str  # type: ignore[misc, unused-ignore]
    uri: Optional[str] = None
    urn: Optional[str] = None


class NameableArtefact(IdentifiableArtefact, frozen=True, omit_defaults=True):
    """Nameable Artefact class.

    Provides a name and a description to all derived classes.

    Attributes:
        name: The name of the artefact.
        description: The description of the artefact.
    """

    name: Optional[str] = None
    description: Optional[str] = None


class VersionableArtefact(NameableArtefact, frozen=True, omit_defaults=True):
    """Versionable Artefact class.

    Provides a version to all derived classes.

    Attributes:
        version: The version of the artefact.
        valid_from: The date from which the artefact is valid.
        valid_to: The date to which the artefact is valid.
    """

    version: str = "1.0"
    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None


class Item(NameableArtefact, frozen=True, omit_defaults=True):
    """Item class.

    The Item is an item of content in an Item Scheme. This may be a
    concept in a concept scheme, a code in a codelist, etc.

    Parent and child attributes (hierarchy) have been removed for simplicity.
    """


class Contact(Struct, frozen=True, omit_defaults=True):
    """Contact details such as the name of a contact and his email address.

    Attributes:
        id: An identifier for a contact. If the contact represents a person,
            this could be the person's username in the organisation.
        name: The contact name, which could be the name of a person, the name
            of a service ("e.g. Support"), etc.
        department: The department in which the contact is located (e.g.
            "Statistics").
        role: The contact's role, which could be his job title, or a role such
            as data owner, data steward, subject matter expert, etc.
        telephones: A list of telephone numbers.
        faxes: A list of fax numbers.
        uris: A list of URLs relevant for the contact (e.g. a link to an online
            form that can be used to send questions, a link to a support forum,
            etc.).
        emails: a list of email addresses.
    """

    id: Optional[str] = None
    name: Optional[str] = None
    department: Optional[str] = None
    role: Optional[str] = None
    telephones: Optional[Sequence[str]] = None
    faxes: Optional[Sequence[str]] = None
    uris: Optional[Sequence[str]] = None
    emails: Optional[Sequence[str]] = None


class Organisation(Item, frozen=True, omit_defaults=True):
    """Organisation class.

    Attributes:
        contacts: The contact of the agency.
        dataflows: The dataflows relevant for the organisation. For example,
            the list of dataflows for which a data provider provides data.
    """

    contacts: Sequence[Contact] = ()
    dataflows: Sequence["DataflowRef"] = ()

    def __hash__(self) -> int:
        """Returns the organisation's hash."""
        return hash(self.id)


class Agency(Organisation, frozen=True, omit_defaults=True):
    """An organisation that maintains structural metadata.

    This includes statistical classifications, glossaries,
    structural metadata such as Data and Metadata Structure
    Definitions, Concepts and Code lists.
    """


class DataProvider(Organisation, frozen=True, omit_defaults=True):
    """An organisation that provides data or metadata."""


class DataConsumer(Organisation, frozen=True, omit_defaults=True):
    """An organisation that collects data or metadata."""


class MaintainableArtefact(
    VersionableArtefact, frozen=True, omit_defaults=True
):
    """Maintainable Artefact class.

    An abstract class to group together primary structural metadata
    artefacts that are maintained by an Agency.

    Attributes:
        is_final: Whether the artefact is final.
        is_external_reference: Whether the artefact is an external reference.
        service_url: The URL of the service.
        structure_url: The URL of the structure.
        agency: The maintainer of the artefact.
    """

    is_final: bool = False
    is_external_reference: bool = False
    service_url: Optional[str] = None
    structure_url: Optional[str] = None
    agency: Union[str, Agency] = ""

    def __post_init__(self) -> None:
        """Additional validation checks for maintainable artefacts."""
        if not self.agency:
            raise Invalid(
                "Missing agency",
                "Maintainable artefacts must reference an agency.",
            )


class ItemScheme(MaintainableArtefact, frozen=True, omit_defaults=True):
    """ItemScheme class.

    The descriptive information for an arrangement or division of objects
    into groups based on characteristics, which the objects have in common.

    Attributes:
        items: The list of items in the scheme.
        is_partial: Whether the scheme is partial.
    """

    items: Sequence[Item] = ()
    is_partial: bool = False


class DataflowRef(MaintainableArtefact, frozen=True, omit_defaults=True):
    """Provide core information about a dataflow.

    Attributes:
        id: The dataflow identifier (e.g. BIS_MACRO).
        agency: The organisation (or unit) responsible for the dataflow.
        name: The dataflow name (e.g. MACRO dataflow).
        description: Additional descriptive information about the dataflow.
        version: The version of the dataflow (e.g. 1.0).
    """



"""SDMX 2.0 CSV writer module."""

from copy import copy
from typing import Optional

import pandas as pd

from pysdmx.io.csv.sdmx20 import SDMX_CSV_ACTION_MAPPER
from pysdmx.io.pd import PandasDataset


def writer(
    dataset: PandasDataset, output_path: Optional[str] = None
) -> Optional[str]:
    """Converts a dataset to an SDMX CSV format.

    Args:
        dataset: dataset
        output_path: output_path

    Returns:
        SDMX CSV data as a string
    """
    # Link to pandas.to_csv documentation on sphinx:
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_csv.html

    # Create a copy of the dataset
    df: pd.DataFrame = copy(dataset.data)

    # Add additional attributes to the dataset
    for k, v in dataset.attributes.items():
        df[k] = v

    structure_ref, unique_id = dataset.short_urn.split("=", maxsplit=1)
    if structure_ref in ["DataStructure", "DataFlow"]:
        structure_ref = structure_ref.lower()
    else:
        structure_ref = "dataprovision"

    # Insert two columns at the beginning of the data set
    df.insert(0, "STRUCTURE", structure_ref)
    df.insert(1, "STRUCTURE_ID", unique_id)
    action_value = SDMX_CSV_ACTION_MAPPER[dataset.action]
    df.insert(2, "ACTION", action_value)

    # Convert the dataset into a csv file
    return df.to_csv(output_path, index=False, header=True)



"""Collection of Fusion-JSON schemas for SDMX-REST schema queries."""

from typing import Dict, List, Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.errors import InternalError
from pysdmx.io.json.fusion.messages.code import FusionCodelist
from pysdmx.io.json.fusion.messages.concept import (
    FusionConcept,
    FusionConceptScheme,
)
from pysdmx.io.json.fusion.messages.constraint import FusionContentConstraint
from pysdmx.io.json.fusion.messages.core import (
    FusionRepresentation,
    FusionString,
)
from pysdmx.model import (
    ArrayBoundaries,
    Codelist,
    Component,
    Components,
    DataType,
    Facets,
    Role,
)
from pysdmx.util import parse_item_urn


def _find_concept(
    cs: Sequence[FusionConceptScheme],
    urn: str,
) -> FusionConcept:
    r = parse_item_urn(urn)
    f = [
        m
        for m in cs
        if (m.agency == r.agency and m.id == r.id and m.version == r.version)
    ]
    return [c for c in f[0].items if c.id == r.item_id][0]


def _get_representation(
    id_: str,
    r: Optional[FusionRepresentation],
    cls: Sequence[FusionCodelist],
    cons: Dict[str, Sequence[str]],
) -> Tuple[
    Optional[DataType],
    Optional[Facets],
    Optional[Codelist],
    Optional[ArrayBoundaries],
]:
    ab = r.to_array_def() if r else None
    dt = DataType(r.textFormat.textType) if r and r.textFormat else None
    facets = r.to_facets() if r else None
    codes = r.to_enumeration(cls, cons.get(id_, [])) if r else None
    return (dt, facets, codes, ab)


class FusionGroup(Struct, frozen=True):
    """Fusion-JSON payload for a group."""

    id: str
    dimensionReferences: Sequence[str]


class FusionAttribute(Struct, frozen=True):
    """Fusion-JSON payload for an attribute."""

    id: str
    concept: str
    mandatory: bool
    attachmentLevel: str
    representation: Optional[FusionRepresentation] = None
    attachmentGroup: Optional[str] = None
    dimensionReferences: Optional[Sequence[str]] = None
    measureReferences: Optional[Sequence[str]] = None

    def __derive_level(self, groups: Sequence[FusionGroup]) -> str:
        if self.attachmentLevel == "OBSERVATION":
            return "O"
        elif self.attachmentLevel == "DATA_SET":
            return "D"
        elif self.attachmentLevel == "GROUP":
            grp = [g for g in groups if g.id == self.attachmentGroup]
            return ",".join(grp[0].dimensionReferences)
        elif self.dimensionReferences:
            return ",".join(self.dimensionReferences)
        else:
            raise InternalError(
                "Invalid metadata",
                (
                    "Could not infer attribute attachment level. "
                    f"The attribute is {self}"
                ),
            )

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[FusionGroup],
    ) -> Component:
        """Returns an attribute."""
        c = _find_concept(cs, self.concept)
        dt, facets, codes, ab = _get_representation(
            self.id, self.representation, cls, cons
        )
        lvl = self.__derive_level(groups)
        if c.descriptions:
            desc = c.descriptions[0].value
        else:
            desc = None
        return Component(
            id=self.id,
            required=self.mandatory,
            role=Role.ATTRIBUTE,
            concept=c.to_model(cls),
            local_dtype=dt,
            local_facets=facets,
            name=c.names[0].value,
            description=desc,
            local_codes=codes,
            attachment_level=lvl,
            array_def=ab,
        )


class FusionAttributes(Struct, frozen=True):
    """Fusion-JSON payload for the list of attributes."""

    attributes: Sequence[FusionAttribute]

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[FusionGroup],
    ) -> List[Component]:
        """Returns the list of attributes."""
        return [d.to_model(cs, cls, cons, groups) for d in self.attributes]


class FusionDimension(Struct, frozen=True):
    """Fusion-JSON payload for a dimension."""

    id: str
    concept: str
    representation: Optional[FusionRepresentation] = None

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a dimension."""
        c = _find_concept(cs, self.concept)
        dt, facets, codes, ab = _get_representation(
            self.id, self.representation, cls, cons
        )
        if c.descriptions:
            desc = c.descriptions[0].value
        else:
            desc = None
        return Component(
            id=self.id,
            required=True,
            role=Role.DIMENSION,
            concept=c.to_model(cls),
            local_dtype=dt,
            local_facets=facets,
            name=c.names[0].value,
            description=desc,
            local_codes=codes,
            array_def=ab,
        )


class FusionDimensions(Struct, frozen=True):
    """Fusion-JSON payload for the list of dimensions."""

    dimensions: Sequence[FusionDimension]

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
    ) -> List[Component]:
        """Returns the list of dimensions."""
        return [d.to_model(cs, cls, cons) for d in self.dimensions]


class FusionMeasure(Struct, frozen=True):
    """Fusion-JSON payload for a measure."""

    id: str
    concept: str
    mandatory: bool
    representation: Optional[FusionRepresentation] = None

    def to_model(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a measure."""
        c = _find_concept(cs, self.concept)
        dt, facets, codes, ab = _get_representation(
            self.id, self.representation, cls, cons
        )
        if c.descriptions:
            desc = c.descriptions[0].value
        else:
            desc = None
        return Component(
            id=self.id,
            required=self.mandatory,
            role=Role.MEASURE,
            concept=c.to_model(cls),
            local_dtype=dt,
            local_facets=facets,
            name=c.names[0].value,
            description=desc,
            local_codes=codes,
            array_def=ab,
        )


class FusionDataStructure(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a DSD."""

    id: str
    names: Sequence[FusionString]
    agency: str
    dimensionList: FusionDimensions
    measures: Sequence[FusionMeasure] = ()
    attributeList: Optional[FusionAttributes] = None
    groups: Sequence[FusionGroup] = ()
    description: Optional[Sequence[FusionString]] = None
    version: str = "1.0"

    def get_components(
        self,
        cs: Sequence[FusionConceptScheme],
        cls: Sequence[FusionCodelist],
        constraints: Sequence[FusionContentConstraint],
    ) -> Components:
        """Returns the schema for this DSD."""
        comps = []
        if constraints:
            cons = constraints[0].to_map()
        else:
            cons = {}
        comps.extend(self.dimensionList.to_model(cs, cls, cons))
        if self.measures:
            comps.extend([m.to_model(cs, cls, cons) for m in self.measures])
        if self.attributeList:
            comps.extend(
                self.attributeList.to_model(
                    cs,
                    cls,
                    cons,
                    self.groups,
                )
            )
        return Components(comps)



"""Collection of SDMX-JSON schemas for SDMX-REST schema queries."""

from datetime import datetime
from typing import Dict, List, Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.concept import (
    JsonConcept,
    JsonConceptScheme,
)
from pysdmx.io.json.sdmxjson2.messages.constraint import JsonDataConstraint
from pysdmx.io.json.sdmxjson2.messages.core import (
    JsonAnnotation,
    JsonRepresentation,
)
from pysdmx.model import (
    ArrayBoundaries,
    Codelist,
    Component,
    Components,
    DataType,
    Facets,
    Role,
)
from pysdmx.util import parse_item_urn


def _find_concept(cs: Sequence[JsonConceptScheme], urn: str) -> JsonConcept:
    r = parse_item_urn(urn)
    f = [
        m
        for m in cs
        if (m.agency == r.agency and m.id == r.id and m.version == r.version)
    ]
    return [c for c in f[0].concepts if c.id == r.item_id][0]


def __get_type(repr_: JsonRepresentation) -> str:
    if repr_.enumerationFormat:
        t = repr_.enumerationFormat.dataType
    elif repr_.format:
        t = repr_.format.dataType
    else:
        t = "String"
    return t


def _get_representation(
    id_: str,
    local: Optional[JsonRepresentation],
    cls: Sequence[Codelist],
    cons: Dict[str, Sequence[str]],
) -> Tuple[
    Optional[DataType],
    Optional[Facets],
    Optional[Codelist],
    Optional[ArrayBoundaries],
]:
    valid = cons.get(id_, [])
    codes = local.to_enumeration(cls, valid) if local else None
    dt = DataType(__get_type(local)) if local else None
    facets = local.to_facets() if local else None
    ab = local.to_array_def() if local else None
    return (dt, facets, codes, ab)


class JsonGroup(Struct, frozen=True):
    """SDMX-JSON payload for a group."""

    id: str
    groupDimensions: Sequence[str]


class JsonAttributeRelationship(Struct, frozen=True):
    """SDMX-JSON payload for an attribute relationship."""

    dataflow: Optional[Dict] = None  # type: ignore[type-arg]
    dimensions: Optional[Sequence[str]] = None
    group: Optional[str] = None
    observation: Optional[Dict] = None  # type: ignore[type-arg]

    def to_model(
        self, groups: Sequence[JsonGroup], measures: Optional[Sequence[str]]
    ) -> str:
        """Returns the attachment level."""
        if measures:
            return "O"
        elif self.dimensions:
            return ",".join(self.dimensions)
        elif self.group:
            grp = [g for g in groups if g.id == self.group]
            return ",".join(grp[0].groupDimensions)
        else:
            return "D"


class JsonDimension(Struct, frozen=True):
    """SDMX-JSON payload for a component."""

    id: str
    conceptIdentity: str
    position: Optional[int] = None
    conceptRoles: Optional[Sequence[str]] = None
    localRepresentation: Optional[JsonRepresentation] = None

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a component."""
        c = _find_concept(cs, self.conceptIdentity)
        dt, facets, codes, ab = _get_representation(
            self.id, self.localRepresentation, cls, cons
        )
        return Component(
            id=self.id,
            required=True,
            role=Role.DIMENSION,
            concept=c.to_model(cls),
            local_dtype=dt,
            local_facets=facets,
            name=c.name,
            description=c.description,
            local_codes=codes,
            array_def=ab,
        )


class JsonAttribute(Struct, frozen=True):
    """SDMX-JSON payload for an attribute."""

    id: str
    conceptIdentity: str
    attributeRelationship: JsonAttributeRelationship
    conceptRoles: Optional[Sequence[str]] = None
    usage: str = "optional"
    measureRelationship: Optional[Sequence[str]] = None
    localRepresentation: Optional[JsonRepresentation] = None

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[JsonGroup],
    ) -> Component:
        """Returns a component."""
        c = _find_concept(cs, self.conceptIdentity)
        dt, facets, codes, ab = _get_representation(
            self.id, self.localRepresentation, cls, cons
        )
        req = self.usage != "optional"
        lvl = self.attributeRelationship.to_model(
            groups,
            self.measureRelationship,
        )
        return Component(
            id=self.id,
            required=req,
            role=Role.ATTRIBUTE,
            concept=c.to_model(cls),
            local_dtype=dt,
            local_facets=facets,
            name=c.name,
            description=c.description,
            local_codes=codes,
            attachment_level=lvl,
            array_def=ab,
        )


class JsonMeasure(Struct, frozen=True):
    """SDMX-JSON payload for a measure."""

    id: str
    conceptIdentity: str
    conceptRoles: Optional[Sequence[str]] = None
    usage: str = "optional"
    localRepresentation: Optional[JsonRepresentation] = None

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> Component:
        """Returns a component."""
        c = _find_concept(cs, self.conceptIdentity)
        dt, facets, codes, ab = _get_representation(
            self.id, self.localRepresentation, cls, cons
        )
        req = self.usage != "optional"
        return Component(
            id=self.id,
            required=req,
            role=Role.MEASURE,
            concept=c.to_model(cls),
            local_dtype=dt,
            local_facets=facets,
            name=c.name,
            description=c.description,
            local_codes=codes,
            array_def=ab,
        )


class JsonAttributes(Struct, frozen=True):
    """SDMX-JSON payload for the list of attributes."""

    attributes: Sequence[JsonAttribute] = ()

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
        groups: Sequence[JsonGroup],
    ) -> List[Component]:
        """Returns the list of attributes."""
        return [d.to_model(cs, cls, cons, groups) for d in self.attributes]


class JsonDimensions(Struct, frozen=True):
    """SDMX-JSON payload for the list of dimensions."""

    dimensions: Sequence[JsonDimension]
    timeDimension: Optional[JsonDimension] = None

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> List[Component]:
        """Returns the list of dimensions."""
        c = []
        c.extend([d.to_model(cs, cls, cons) for d in self.dimensions])
        if self.timeDimension:
            c.append(self.timeDimension.to_model(cs, cls, cons))
        return c


class JsonMeasures(Struct, frozen=True):
    """SDMX-JSON payload for the list of measures."""

    measures: Sequence[JsonMeasure]

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        cons: Dict[str, Sequence[str]],
    ) -> List[Component]:
        """Returns the list of measures."""
        return [m.to_model(cs, cls, cons) for m in self.measures]


class JsonComponents(Struct, frozen=True):
    """SDMX-JSON payload for the list of DSD components."""

    dimensionList: JsonDimensions
    measureList: Optional[JsonMeasures] = None
    attributeList: Optional[JsonAttributes] = None
    groups: Sequence[JsonGroup] = ()

    def to_model(
        self,
        cs: Sequence[JsonConceptScheme],
        cls: Sequence[Codelist],
        constraints: Sequence[JsonDataConstraint],
    ) -> Components:
        """Returns the schema for this DSD."""
        comps = []
        if constraints and constraints[0].cubeRegions:
            cons = constraints[0].cubeRegions[0].to_map()
        else:
            cons = {}
        comps.extend(self.dimensionList.to_model(cs, cls, cons))
        if self.measureList:
            comps.extend(self.measureList.to_model(cs, cls, cons))
        if self.attributeList:
            comps.extend(
                self.attributeList.to_model(
                    cs,
                    cls,
                    cons,
                    self.groups,
                )
            )
        return Components(comps)


class JsonDataStructure(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a DSD."""

    id: str
    name: str
    agency: str
    dataStructureComponents: JsonComponents
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    metadata: Optional[str] = None



"""SDMX 2.1 writer package."""

from typing import Any, Dict, Optional

from pysdmx.errors import NotImplemented
from pysdmx.io.xml.enums import MessageType
from pysdmx.io.xml.sdmx21.writer.__write_aux import (
    __write_header,
    create_namespaces,
    get_end_message,
)
from pysdmx.io.xml.sdmx21.writer.structure import (
    generate_structures,
)
from pysdmx.model.message import Header


def writer(
    content: Dict[str, Any],
    type_: MessageType,
    path: str = "",
    prettyprint: bool = True,
    header: Optional[Header] = None,
) -> Optional[str]:
    """This function writes a SDMX-ML file from the Message Content.

    Args:
        content: The content to be written
        type_: The type of message to be written
        path: The path to save the file
        prettyprint: Prettyprint or not
        header: The header to be used (generated if None)

    Returns:
        The XML string if path is empty, None otherwise

    Raises:
        NotImplemented: If the MessageType is not Metadata
    """
    if type_ != MessageType.Structure:
        raise NotImplemented(
            "Unsupported", "Only Metadata messages are supported"
        )
    outfile = create_namespaces(type_, content, prettyprint)

    if header is None:
        header = Header()

    outfile += __write_header(header, prettyprint)

    outfile += generate_structures(content, prettyprint)

    outfile += get_end_message(type_, prettyprint)

    if path == "":
        return outfile

    with open(path, "w", encoding="UTF-8", errors="replace") as f:
        f.write(outfile)

    return None



"""Parsers for reading metadata."""

from typing import Any, Dict

from msgspec import Struct

from pysdmx.io.xml.sdmx21.__parsing_config import CORE_REP, URN
from pysdmx.io.xml.sdmx21.reader.__utils import (
    AGENCIES,
    AGENCY,
    AGENCY_ID,
    ANNOTATION,
    ANNOTATION_TEXT,
    ANNOTATION_TITLE,
    ANNOTATION_TYPE,
    ANNOTATION_URL,
    ANNOTATIONS,
    CL,
    CLS,
    CODE,
    CON,
    CONTACT,
    CS,
    DEPARTMENT,
    DESC,
    EMAIL,
    EMAILS,
    FACETS,
    FacetType,
    FAX,
    FAXES,
    ID,
    IS_EXTERNAL_REF,
    IS_EXTERNAL_REF_LOW,
    IS_FINAL,
    IS_FINAL_LOW,
    IS_PARTIAL,
    IS_PARTIAL_LOW,
    NAME,
    ROLE,
    SER_URL,
    SER_URL_LOW,
    STR_URL,
    STR_URL_LOW,
    TELEPHONE,
    TELEPHONES,
    TEXT,
    TEXT_TYPE,
    TITLE,
    TYPE,
    unique_id,
    URI,
    URIS,
    URL,
    VERSION,
)
from pysdmx.io.xml.utils import add_list
from pysdmx.model import (
    Code,
    Codelist,
    Concept,
    ConceptScheme,
    DataType,
    Facets,
)
from pysdmx.model.__base import Agency, Annotation, Contact, Item, ItemScheme
from pysdmx.model.message import CONCEPTS, ORGS
from pysdmx.util import find_by_urn

SCHEMES_CLASSES = {CL: Codelist, AGENCIES: ItemScheme, CS: ConceptScheme}
ITEMS_CLASSES = {AGENCY: Agency, CODE: Code, CON: Concept}


class StructureParser(Struct):
    """StructureParser class for SDMX-ML 2.1."""

    agencies: Dict[str, Any] = {}
    codelists: Dict[str, Any] = {}
    concepts: Dict[str, Any] = {}
    datastructures: Dict[str, Any] = {}
    dataflows: Dict[str, Any] = {}

    def __extract_text(self, element: Any) -> str:
        """Extracts the text from the element.

        Args:
            element: The element to extract the text from

        Returns:
            The text extracted from the element
        """
        if isinstance(element, dict) and "#text" in element:
            element = element["#text"]
        return element

    def __format_contact(self, json_contact: Dict[str, Any]) -> Contact:
        """Creates a Contact object from a json_contact.

        Args:
            json_contact: The element to create the Contact object from

        Returns:
            Contact object created from the json_contact
        """
        self.__format_name_description(json_contact)

        xml_node_to_attribute = {
            NAME: NAME.lower(),
            DEPARTMENT: DEPARTMENT.lower(),
            ROLE: ROLE.lower(),
            URI: URIS,
            EMAIL: EMAILS,
            TELEPHONE: TELEPHONES,
            FAX: FAXES,
        }

        for k, v in xml_node_to_attribute.items():
            if k in json_contact:
                if k in [DEPARTMENT, ROLE]:
                    json_contact[v] = self.__extract_text(json_contact.pop(k))
                    continue
                field_info = add_list(json_contact.pop(k))
                for i, element in enumerate(field_info):
                    field_info[i] = self.__extract_text(element)
                json_contact[v] = field_info

        return Contact(**json_contact)

    def __format_annotations(self, item_elem: Any) -> Dict[str, Any]:
        """Formats the annotations in the item_elem.

        Args:
            item_elem: The element to be formatted

        Returns:
            annotations formatted
        """
        if ANNOTATIONS not in item_elem:
            return item_elem
        annotations = []

        ann = item_elem[ANNOTATIONS]
        ann[ANNOTATION] = add_list(ann[ANNOTATION])
        for e in ann[ANNOTATION]:
            if ANNOTATION_TITLE in e:
                e[TITLE] = e.pop(ANNOTATION_TITLE)
            if ANNOTATION_TYPE in e:
                e[TYPE] = e.pop(ANNOTATION_TYPE)
            if ANNOTATION_TEXT in e:
                e[TEXT] = self.__extract_text(e[ANNOTATION_TEXT])
                del e[ANNOTATION_TEXT]
            if ANNOTATION_URL in e:
                e[URL] = e.pop(ANNOTATION_URL)

            annotations.append(Annotation(**e))

        item_elem[ANNOTATIONS.lower()] = annotations
        del item_elem[ANNOTATIONS]

        return item_elem

    def __format_name_description(self, element: Any) -> Dict[str, Any]:
        node = [NAME, DESC]
        for field in node:
            if field in element:
                element[field.lower()] = self.__extract_text(element[field])
                del element[field]
        return element

    @staticmethod
    def __format_facets(
        json_fac: Dict[str, Any], json_obj: Dict[str, Any]
    ) -> None:
        """Formats the facets in the json_fac to be stored in json_obj.

        Args:
            json_fac: The element with the facets to be formatted
            json_obj: The element to store the formatted facets
        """
        for key, _value in json_fac.items():
            if key == TEXT_TYPE and json_fac[TEXT_TYPE] in list(DataType):
                json_obj["dtype"] = DataType(json_fac[TEXT_TYPE])

            if key in FacetType:
                facet_kwargs = {
                    FacetType[k]: v
                    for k, v in json_fac.items()
                    if k in FacetType
                }
                json_obj[FACETS] = Facets(**facet_kwargs)

    @staticmethod
    def __format_urls(json_elem: Dict[str, Any]) -> Dict[str, Any]:
        """Formats the STR_URL and SER_URL keys in the element.

        Args:
            json_elem: The element to be formatted

        Returns:
            The json_elem with STR_URL and SER_URL keys formatted.
        """
        if STR_URL in json_elem:
            json_elem[STR_URL_LOW] = json_elem.pop(STR_URL)
        if SER_URL in json_elem:
            json_elem[SER_URL_LOW] = json_elem.pop(SER_URL)
        return json_elem

    def __format_agency(self, element: Dict[str, Any]) -> Dict[str, Any]:
        """Formats the AGENCY_ID key in the element to the maintainer.

        Args:
            element: The element with the Agency ID to be formatted

        Returns:
            element with the Agency ID formatted
        """
        element[AGENCY.lower()] = self.agencies.get(
            element[AGENCY_ID], element[AGENCY_ID]
        )
        del element[AGENCY_ID]
        return element

    def __format_orgs(self, json_orgs: Dict[str, Any]) -> Dict[str, Any]:
        orgs: Dict[str, Any] = {}
        json_orgs = add_list(json_orgs)  # type: ignore[assignment]
        for e in json_orgs:
            ag_sch = self.__format_scheme(
                e, AGENCIES, AGENCY  # type: ignore[arg-type]
            )
            orgs = {**orgs, **ag_sch}
        return orgs

    def __format_representation(
        self, json_rep: Dict[str, Any], json_obj: Dict[str, Any]
    ) -> None:
        """Formats the representation in the json_rep."""
        if "TextFormat" in json_rep:
            self.__format_facets(json_rep["TextFormat"], json_obj)

        if (
            "Enumeration" in json_rep
            and URN in json_rep["Enumeration"]
            and len(self.codelists) > 0
        ):
            codelist = find_by_urn(
                list(self.codelists.values()),
                json_rep["Enumeration"][URN],
            )
            json_obj["codes"] = codelist.codes

    def __format_item(
        self, item_json_info: Dict[str, Any], item_name_class: str
    ) -> Item:
        item_json_info = self.__format_annotations(item_json_info)
        item_json_info = self.__format_name_description(item_json_info)
        if CONTACT in item_json_info and item_name_class == AGENCY:
            item_json_info[CONTACT] = add_list(item_json_info[CONTACT])
            contacts = []
            for e in item_json_info[CONTACT]:
                contacts.append(self.__format_contact(e))
            item_json_info[CONTACT.lower() + "s"] = contacts
            del item_json_info[CONTACT]

        if CORE_REP in item_json_info and item_name_class == CON:
            self.__format_representation(
                item_json_info[CORE_REP], item_json_info
            )
            del item_json_info[CORE_REP]

        if "Parent" in item_json_info:
            del item_json_info["Parent"]

        return ITEMS_CLASSES[item_name_class](**item_json_info)

    def __format_scheme(
        self, json_elem: Dict[str, Any], scheme: str, item: str
    ) -> Dict[str, Any]:
        elements: Dict[str, Any] = {}

        json_elem[scheme] = add_list(json_elem[scheme])
        for element in json_elem[scheme]:
            element["items"] = []

            element = self.__format_annotations(element)
            element = self.__format_name_description(element)
            full_id = unique_id(
                element[AGENCY_ID], element[ID], element[VERSION]
            )
            element = self.__format_urls(element)
            if IS_EXTERNAL_REF in element:
                element[IS_EXTERNAL_REF_LOW] = element.pop(IS_EXTERNAL_REF)
            if IS_FINAL in element:
                element[IS_FINAL_LOW] = element.pop(IS_FINAL)
            if IS_PARTIAL in element:
                element[IS_PARTIAL_LOW] = element.pop(IS_PARTIAL)
            element[item] = add_list(element[item])
            items = []
            for item_elem in element[item]:
                # Dynamic
                items.append(self.__format_item(item_elem, item))
            del element[item]
            element["items"] = items
            if scheme == AGENCIES:
                self.agencies.update({e.id: e for e in items})
            if scheme == CS:
                self.concepts.update({e.id: e for e in items})
            element = self.__format_agency(element)
            # Dynamic creation with specific class
            elements[full_id] = SCHEMES_CLASSES[scheme](**element)

        return elements

    def format_structures(self, json_meta: Dict[str, Any]) -> Dict[str, Any]:
        """Formats the structures in json format.

        Args:
            json_meta: The structures in json format

        Returns:
            A dictionary with the structures formatted
        """
        # Reset dict to store metadata
        structures = {}

        if ORGS in json_meta:
            structures[ORGS] = self.__format_orgs(json_meta[ORGS])
        if CLS in json_meta:
            structures[CLS] = self.__format_scheme(json_meta[CLS], CL, CODE)
            self.codelists = structures[CLS]
        if CONCEPTS in json_meta:
            structures[CONCEPTS] = self.__format_scheme(
                json_meta[CONCEPTS], CS, CON
            )
        # Reset global variables
        return structures



"""SDMX 1.0 CSV writer module."""

from copy import copy
from typing import Optional

import pandas as pd

from pysdmx.io.pd import PandasDataset


def writer(
    dataset: PandasDataset, output_path: Optional[str] = None
) -> Optional[str]:
    """Converts a dataset to an SDMX CSV format.

    Args:
        dataset: dataset
        output_path: output_path

    Returns:
        SDMX CSV data as a string
    """
    # Link to pandas.to_csv documentation on sphinx:
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_csv.html

    # Create a copy of the dataset
    df: pd.DataFrame = copy(dataset.data)
    df.insert(0, "DATAFLOW", dataset.short_urn.split("=")[1])

    # Add additional attributes to the dataset
    for k, v in dataset.attributes.items():
        df[k] = v

    # Return the SDMX CSV data as a string
    return df.to_csv(output_path, index=False, header=True)



"""Collection of Fusion-JSON schemas for SDMX-REST schema queries."""

from typing import List, Sequence

import msgspec

from pysdmx.io.json.fusion.messages.code import FusionCodelist
from pysdmx.io.json.fusion.messages.concept import FusionConceptScheme
from pysdmx.io.json.fusion.messages.constraint import FusionContentConstraint
from pysdmx.io.json.fusion.messages.core import FusionLink
from pysdmx.io.json.fusion.messages.dsd import FusionDataStructure
from pysdmx.model import Components, HierarchyAssociation, Schema
from pysdmx.util import parse_item_urn


class FusionHeader(msgspec.Struct, frozen=True):
    """Fusion-JSON payload for message header."""

    links: Sequence[FusionLink] = ()


class FusionSchemaMessage(msgspec.Struct, frozen=True):
    """Fusion-JSON payload for /schema queries."""

    meta: FusionHeader
    ConceptScheme: Sequence[FusionConceptScheme]
    DataStructure: Sequence[FusionDataStructure]
    ValueList: Sequence[FusionCodelist] = ()
    Codelist: Sequence[FusionCodelist] = ()
    DataConstraint: Sequence[FusionContentConstraint] = ()

    def to_model(
        self,
        context: str,
        agency: str,
        id_: str,
        version: str,
        hierarchies: Sequence[HierarchyAssociation],
    ) -> Schema:
        """Returns the requested schema."""
        cls: List[FusionCodelist] = []
        cls.extend(self.Codelist)
        cls.extend(self.ValueList)
        components = self.DataStructure[0].get_components(
            self.ConceptScheme, cls, self.DataConstraint
        )
        comp_dict = {c.id: c for c in components}
        urns = [a.urn for a in self.meta.links]
        for ha in hierarchies:
            comp_id = parse_item_urn(ha.component_ref).item_id
            h = msgspec.structs.replace(  # type: ignore[type-var]
                ha.hierarchy,
                operator=ha.operator,
            )
            comp_dict[comp_id] = msgspec.structs.replace(
                components[comp_id], local_codes=h
            )
            urns.append(
                "urn:sdmx:org.sdmx.infomodel.codelist.Hierarchy="
                f"{h.agency}:{h.id}({h.version})"  # type: ignore[union-attr]
            )
        comps = Components(comp_dict.values())
        return Schema(context, agency, id_, comps, version, urns)



"""Utility classes for pysdmx REST query builders."""

from enum import IntEnum
import re
from typing import Sequence, Union

from pysdmx.errors import Invalid


class ApiVersion(IntEnum):
    """The version of the SDMX-REST API."""

    V1_0_0 = 0
    V1_0_1 = 1
    V1_0_2 = 2
    V1_1_0 = 3
    V1_2_0 = 4
    V1_3_0 = 5
    V1_4_0 = 6
    V1_5_0 = 7
    V2_0_0 = 8
    V2_1_0 = 9


MULT_SEP = re.compile(r"\+")
REST_ALL = "*"
REST_LATEST = "~"


def check_multiple_items(
    value: Union[str, Sequence[str]], version: ApiVersion
) -> None:
    """Whether multiple items are supported in the supplied API version."""
    if not isinstance(value, str) and version < ApiVersion.V1_3_0:
        raise Invalid(
            "Validation Error",
            f"Multiple items not allowed in SDMX-REST {version.value}.",
        )


def check_multiple_data_context(
    field: str, value: Union[str, Sequence[str]], version: ApiVersion
) -> None:
    """Whether multiple items are supported in the supplied API version."""
    if not isinstance(value, str) and version < ApiVersion.V2_0_0:
        raise Invalid(
            "Validation Error",
            (
                f"More than one {field} is not allowed in data context "
                f"for SDMX-REST {version.value}."
            ),
        )


__all__ = ["ApiVersion", "check_multiple_items", "check_multiple_data_context"]



"""SDMX 2.1 XML reader package."""

from typing import Any, Dict, Optional

import xmltodict

from pysdmx.errors import Invalid, NotImplemented
from pysdmx.io.xml.enums import MessageType
from pysdmx.io.xml.sdmx21.__parsing_config import (
    DATASET,
    ERROR,
    ERROR_CODE,
    ERROR_MESSAGE,
    ERROR_TEXT,
    GENERIC,
    HEADER,
    REG_INTERFACE,
    STRSPE,
    STRUCTURE,
    STRUCTURES,
    XML_OPTIONS,
)
from pysdmx.io.xml.sdmx21.doc_validation import validate_doc
from pysdmx.io.xml.sdmx21.reader.data_read import (
    __extract_structure,
    create_dataset,
)
from pysdmx.io.xml.sdmx21.reader.metadata_read import StructureParser
from pysdmx.io.xml.sdmx21.reader.submission_reader import (
    handle_registry_interface,
)
from pysdmx.io.xml.utils import add_list

MODES = {
    MessageType.GenericDataSet.value: GENERIC,
    MessageType.StructureSpecificDataSet.value: STRSPE,
    MessageType.Structure.value: STRUCTURE,
    MessageType.Submission.value: REG_INTERFACE,
    MessageType.Error.value: ERROR,
}


def read_xml(
    infile: str,
    validate: bool = True,
    mode: Optional[MessageType] = None,
    use_dataset_id: bool = False,
) -> Dict[str, Any]:
    """Reads an SDMX-ML file and returns a dictionary with the parsed data.

    Args:
        infile: Path to file, URL, or string.
        validate: If True, the XML data will be validated against the XSD.
        mode: The type of message to parse.
        use_dataset_id: If True, the dataset ID will be used as the key in the
            resulting dictionary.

    Returns:
        dict: Dictionary with the parsed data.

    Raises:
        Invalid: If the SDMX data cannot be parsed.
    """
    if validate:
        validate_doc(infile)
    dict_info = xmltodict.parse(
        infile, **XML_OPTIONS  # type: ignore[arg-type]
    )

    del infile

    if mode is not None and MODES[mode.value] not in dict_info:
        raise Invalid(
            "Validation Error",
            f"Unable to parse sdmx file as {MODES[mode.value]} file",
        )

    result = __generate_sdmx_objects_from_xml(dict_info, use_dataset_id)

    return result


def __generate_sdmx_objects_from_xml(
    dict_info: Dict[str, Any], use_dataset_id: bool = False
) -> Dict[str, Any]:
    """Generates SDMX objects from the XML dictionary (xmltodict).

    Args:
        dict_info: XML dictionary (xmltodict)
        use_dataset_id: Use the dataset ID as the key in
            the resulting dictionary

    Returns:
        dict: Dictionary with the parsed data.

    Raises:
        Invalid: If a SOAP error message is found.
        NotImplemented: If the SDMX data cannot be parsed.
    """
    if ERROR in dict_info:
        code = dict_info[ERROR][ERROR_MESSAGE][ERROR_CODE]
        text = dict_info[ERROR][ERROR_MESSAGE][ERROR_TEXT]
        raise Invalid("Invalid", f"{code}: {text}")
    if STRSPE in dict_info:
        return __parse_dataset(dict_info[STRSPE], mode=STRSPE)
    if GENERIC in dict_info:
        return __parse_dataset(dict_info[GENERIC], mode=GENERIC)
    if STRUCTURE in dict_info:
        return StructureParser().format_structures(
            dict_info[STRUCTURE][STRUCTURES]
        )
    if REG_INTERFACE in dict_info:
        return handle_registry_interface(dict_info)
    raise NotImplemented("Unsupported", "Cannot parse input as SDMX.")


def __parse_dataset(message_info: Dict[str, Any], mode: str) -> Dict[str, Any]:
    """Parse dataset.

    Args:
        message_info: Dict.
        mode: Str.

    Returns:
        A dictionary of datasets.
    """
    str_info = __extract_structure(message_info[HEADER][STRUCTURE])
    dataset_info = add_list(message_info[DATASET])
    datasets = {}
    for dataset in dataset_info:
        ds = create_dataset(dataset, str_info, mode)
        datasets[ds.short_urn] = ds
    return datasets



"""Collection of utility functions."""

import re
from typing import Any, Sequence

from msgspec import Struct

from pysdmx.errors import Invalid, NotFound
from pysdmx.model import Agency
from pysdmx.util._date_pattern_map import convert_dpm

NF = "Not found"


class Reference(Struct, frozen=True):
    """The coordinates of an SDMX maintainable artefact.

    Attributes:
        sdmx_type: The type of SDMX artefact (``codelist``, etc.)
        agency: The maintainer of the artefact (e.g. ``BIS``, ``SDMX``, etc.)
        id: The artefact ID (e.g. ``CL_FREQ``)
        version: The artefact version (e.g. ``1.0.0``)
    """

    sdmx_type: str
    agency: str
    id: str
    version: str

    def __str__(self) -> str:
        """Returns a string representation of the object."""
        return f"{self.sdmx_type}={self.agency}:{self.id}({self.version})"


class ItemReference(Struct, frozen=True):
    """The coordinates of an SDMX non-nested item.

    Attributes:
        sdmx_type: The type of SDMX artefact (``concept``, etc.)
        agency: The maintainer of the artefact (e.g. ``BIS``, etc.)
        id: The maintainable ID (e.g. ``CL_FREQ``)
        version: The artefact version (e.g. ``1.0.0``)
        item_id: The item ID (e.g. ``A``)
    """

    sdmx_type: str
    agency: str
    id: str
    version: str
    item_id: str


maintainable_urn_pattern = re.compile(r"^.*\.(.*)=(.*):(.*)\((.*)\)$")
item_urn_pattern = re.compile(r"^.*\.(.*)=(.*):(.*)\((.*)\)\.(.*)$")


def parse_urn(urn: str) -> Reference:
    """Parses an SDMX urn and returns an object with the details."""
    m = re.match(maintainable_urn_pattern, urn)
    if m:
        return Reference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
        )
    else:
        raise Invalid(NF, f"{urn} does not match {maintainable_urn_pattern}")


def parse_item_urn(urn: str) -> ItemReference:
    """Parses an SDMX item urn and returns an object with the details."""
    m = re.match(item_urn_pattern, urn)
    if m:
        return ItemReference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
            item_id=m.group(5),
        )
    else:
        raise Invalid(NF, f"{urn} does not match {item_urn_pattern}.")


def find_by_urn(artefacts: Sequence[Any], urn: str) -> Any:
    """Returns the maintainable artefact matching the supplied urn."""
    r = parse_urn(urn)
    f = [
        a
        for a in artefacts
        if (
            (
                a.agency == r.agency
                or (
                    a.agency.id == r.agency
                    if isinstance(a.agency, Agency)
                    else False
                )
            )
            and a.id == r.id
            and a.version == r.version
        )
    ]
    if f:
        return f[0]
    else:
        urns = [f"{a.agency}:{a.id}({a.version})" for a in artefacts]
        raise NotFound(
            NF,
            (
                f"Could not find an artefact matching the following URN: "
                f"{urn}. The artefacts received were: {urns}."
            ),
        )


__all__ = ["convert_dpm", "find_by_urn", "parse_item_urn", "parse_urn"]



from enum import Enum

from parsy import from_enum, regex, seq, string  # type: ignore[import-untyped]

from pysdmx.api.dc.query._model import Operator
from pysdmx.api.dc.query._parsing_model import _Filter
from pysdmx.api.dc.query._parsing_util import (
    _field,
    _in_vals,
    _pad,
    _to_response,
    _value,
)


class _PythonCoreOperator(Enum):
    EQUALS = "=="
    NOT_EQUALS = "!="
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN_OR_EQUAL = ">="


class _PythonInOperator(Enum):
    IN = "in"
    NOT_IN = "not in"


class PythonOperator(Enum):
    EQUALS = "=="
    NOT_EQUALS = "!="
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN_OR_EQUAL = ">="
    IN = "in"
    NOT_IN = "not in"


def _to_operator(op: PythonOperator) -> Operator:
    return Operator[op.name]


__lparen = regex(r"[\[\(]{1}")
__rparen = regex(r"[\]\)]{1}")

__co_op = from_enum(_PythonCoreOperator)
__in_op = regex(r"(not)?\s*(in){1}").map(_PythonInOperator)

__core_parser = seq(
    field=_pad >> _field << _pad,
    operator=__co_op.map(_to_operator),
    value=_pad >> _value << _pad,
).combine_dict(_Filter)

__in_parser = seq(
    field=_pad >> _field << _pad,
    operator=__in_op.map(_to_operator),
    value=_pad >> __lparen >> _in_vals << __rparen << _pad,
).combine_dict(_Filter)


__p = __core_parser | __in_parser
py_parser = __p.sep_by(_pad + string("and") + _pad, min=1).map(_to_response)

__all__ = ["py_parser"]



"""Parsing configuration for SDMX-ML 2.1 messages."""

SCHEMA_ROOT = "http://www.sdmx.org/resources/sdmxml/schemas/v2_1/"
NAMESPACES_21 = {
    SCHEMA_ROOT + "message": None,
    SCHEMA_ROOT + "common": None,
    SCHEMA_ROOT + "structure": None,
    "http://www.w3.org/2001/XMLSchema-instance": "xsi",
    "http://www.w3.org/XML/1998/namespace": None,
    SCHEMA_ROOT + "data/structurespecific": None,
    SCHEMA_ROOT + "data/generic": None,
    SCHEMA_ROOT + "registry": None,
    "http://schemas.xmlsoap.org/soap/envelope/": None,
}

XML_OPTIONS = {
    "process_namespaces": True,
    "namespaces": NAMESPACES_21,
    "dict_constructor": dict,
    "attr_prefix": "",
}

# Common
HEADER = "Header"
DATASET = "DataSet"
SERIES = "Series"
OBS = "Obs"
AGENCY_ID = "agencyID"
ID = "id"
VERSION = "version"

# Structure Specific
STRSPE = "StructureSpecificData"
GENERIC = "GenericData"
SERIESKEY = "SeriesKey"
ATTRIBUTES = "Attributes"
VALUE = "Value"
OBS_DIM = "ObsDimension"
OBSVALUE = "ObsValue"
OBSKEY = "ObsKey"
# Header
STRREF = "structureRef"
STRUCTURE = "Structure"
STR_USAGE = "StructureUsage"
STRID = "structureID"
STRTYPE = "structure_type"
DIM_OBS = "dimensionAtObservation"
ALL_DIM = "AllDimensions"
REF = "Ref"
DATASET_ID = "DataSetID"

# SDMX Error handling
ERROR = "Error"
ERROR_MESSAGE = "ErrorMessage"
ERROR_CODE = "code"
ERROR_TEXT = "Text"

# SDMX Registry Interface handling
REG_INTERFACE = "RegistryInterface"
SUBMIT_STRUCTURE_RESPONSE = "SubmitStructureResponse"
SUBMISSION_RESULT = "SubmissionResult"
SUBMITTED_STRUCTURE = "SubmittedStructure"
MAINTAINABLE_OBJECT = "MaintainableObject"
ACTION = "action"
STATUS_MSG = "StatusMessage"
STATUS = "status"

# SOAP API handling
FAULT = "Fault"
FAULTCODE = "faultcode"
FAULTSTRING = "faultstring"


# Structures
# Common
NAME = "Name"
DESC = "Description"
LANG = "lang"
XML_TEXT = "#text"
STR_URL = "structureURL"
STR_URL_LOW = "structureUrl"
SER_URL = "serviceURL"
SER_URL_LOW = "serviceUrl"
# General
ANNOTATIONS = "Annotations"
STRUCTURES = "Structures"
ORGS = "OrganisationSchemes"
AGENCIES = "AgencyScheme"
CODELISTS = "Codelists"
CONCEPTS = "Concepts"
DSDS = "DataStructures"
DATAFLOWS = "Dataflows"
CONSTRAINTS = "Constraints"

# Individual
AGENCY = "Agency"
CL = "Codelist"
CODE = "Code"
CS = "ConceptScheme"
CS_LOW = "concept_scheme"
CON = "Concept"
DSD = "DataStructure"

# DSD components
DSD_COMPS = "DataStructureComponents"
CON_ID = "ConceptIdentity"
CON_ID_LOW = "concept_identity"
CON_ROLE = "ConceptRole"
CON_ROLE_LOW = "concept_role"
# Dimension
DIM_LIST = "DimensionList"
DIM_LIST_LOW = "dimension_list"
DIM = "Dimension"
TIME_DIM = "TimeDimension"
# Attribute
ATT_LIST = "AttributeList"
ATT_LIST_LOW = "attribute_list"
ATT = "Attribute"
ATT_REL = "AttributeRelationship"
AS_STATUS = "assignmentStatus"
# Measure
ME_LIST = "MeasureList"
ME_LIST_LOW = "measure_list"
MEASURE = "Measure"
PRIM_MEASURE = "PrimaryMeasure"
# Group Dimension
GROUP = "Group"
GROUP_DIM_LOW = "group_dimension_descriptor"
GROUP_DIM = "GroupDimension"
DIM_REF = "DimensionReference"

# Dataflows
DF = "Dataflow"

# Constraints
CON_CONS = "ContentConstraint"
CONS_ATT = "ConstraintAttachment"
CUBE_REGION = "CubeRegion"
CONTENT_REGION = "dataContentRegion"
KEY = "Key"
KEY_VALUE = "KeyValue"
DATA_KEY_SET = "DataKeySet"
DATA_KEY_SET_LOW = "dataKeySet"
INCLUDED = "isIncluded"
INCLUDE = "include"

# Annotation
ANNOTATION = "Annotation"
ANNOTATION_TITLE = "AnnotationTitle"
ANNOTATION_TYPE = "AnnotationType"
ANNOTATION_TEXT = "AnnotationText"
ANNOTATION_URL = "AnnotationURL"

TITLE = "title"
TYPE_ = "type_"
TYPE = "type"
TEXT = "text"
URL = "url"
URN = "URN"

# Representation
CORE_REP = "CoreRepresentation"
CORE_REP_LOW = "core_representation"
LOCAL_REP = "LocalRepresentation"
LOCAL_REP_LOW = "local_representation"
ENUM = "Enumeration"
ENUM_FORMAT = "EnumerationFormat"
TEXT_FORMAT = "TextFormat"

# Facets
FACETS = "facets"
TEXT_TYPE = "textType"
TEXT_TYPE_LOW = "text_type"

# Contact
CONTACT = "Contact"
DEPARTMENT = "Department"
ROLE = "Role"
URI = "URI"
EMAIL = "Email"
X400 = "X400"
TELEPHONE = "Telephone"
FAX = "Fax"

# Extras
MAINTAINER = "maintainer"
XMLNS = "xmlns"
COMPS = "components"
PARENT = "Parent"
PAR_ID = "maintainableParentID"
PAR_VER = "maintainableParentVersion"
REL_TO = "relatedTo"
NO_REL = "NoSpecifiedRelationship"

# To exclude from attached_attributes
exc_attributes = [STRREF, "action", "dataScope", "xsi:type", SERIES, OBS]

# Content types
DATASTRUCTURES_CM = "DataStructures"
DATAFLOWS_CM = "Dataflows"
CODELISTS_CM = "Codelists"
CONCEPTS_CM = "Concepts"
ORGANISATIONSCHEMES_CM = "OrganisationSchemes"



"""Model for Mapping Definitions."""

from datetime import datetime
from re import Pattern
from typing import Any, Iterator, Literal, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.model.__base import MaintainableArtefact
from pysdmx.model.concept import DataType
from pysdmx.util import convert_dpm


class DatePatternMap(Struct, frozen=True, omit_defaults=True):
    """A mapping based on a date pattern.

    Examples:
        For example, let's assume a component in the source, with ID `DATE`,
        and containing string values such as `Sep 23`. In the target, we want
        this to be mapped to the `TIME_PERIOD` component, with ISO 8601
        periods for monthly data (e.g. `2023-09`). This can be expressed with
        the following mapping:

            >>> DatePatternMap("DATE", "TIME_PERIOD", "MMM yy", "M")


    Attributes:
        source: The ID of the source component.
        target: The ID of the target component.
        pattern: Describes the source date using conventions for describing
            years, months, days, etc
        frequency: The frequency to convert the input date into or a
            reference to a dimension or an atttribute with the frequency code.
            See `pattern_type` below for additional information.
        id: The Map ID, as defined in the Registry.
        locale: The locale on which the input will be parsed according to the
            pattern.
        pattern_type: The type of date pattern, i.e. fixed or variable. When
            the type is `fixed`, `frequency` is a fixed value from the
            frequency codelist (e.g. `A` for annual frequency). When the type
            is `variable`, `frequency` references a dimension or attribute in
            the target structure (e.g. `FREQ`). In this case, the input date
            can be converted to a different format, depending on the
            frequency of the converted data.
    """

    source: str
    target: str
    pattern: str
    frequency: str
    id: Optional[str] = None
    locale: str = "en"
    pattern_type: Literal["fixed", "variable"] = "fixed"

    @property
    def py_pattern(self) -> str:
        """Translate SDMX pattern into Python format codes."""
        return convert_dpm(self.pattern)


class FixedValueMap(Struct, frozen=True, omit_defaults=True):
    """Set a component to a fixed value.

    Examples:
        For example, let's assume that all observations in the target must be
        treated as free for publication. This can be expressed with the
        following mapping:

            >>> FixedValueMap("CONF_STATUS", "F")

    Attributes:
        target: The ID of the component to which the fixed value is assigned.
        value: The fixed value of the referenced component.
        located_in: Whether the component with a fixed value is in the source
            structure or the target structure. It usually is in the target
            structure (the default), but it can also be in the source, in case
            of bi-directional mapping.
    """

    target: str
    value: Any
    located_in: Literal["source", "target"] = "target"


class ImplicitComponentMap(Struct, frozen=True, omit_defaults=True):
    """A mapping where the value in the source is copied to the target.

    Examples:
        For example, let's assume a component in the source (`OBS_CONF`),
        indicating the confidentiality of data, that we want to map as-is to
        a target component (`CONF_STATUS`). This can be expressed with the
        following mapping:

        >>> ImplicitComponentMap("OBS_CONF", "CONF_STATUS")

    Attributes:
        source:
            The ID of the source component to be mapped, from which
            we want to copy the value.
        target: The ID of the target component.
    """

    source: str
    target: str


class MultiValueMap(Struct, frozen=True, omit_defaults=True):
    """Provides the values for a mapping between one or more components.

    Examples:
        For example, let's assume that we want to map the the code for local
        currency (say `LC`) to an ISO 3-letter currency code, depending on the
        country. So, if the country is `DE` (Germany), then the currency in
        the target should be `EUR` but, if the country is `CH`, then it should
        be `CHF`. This can be expressed with the following value maps:

        >>> MultiValueMap(["DE", "LC"], ["EUR"])
        >>> MultiValueMap(["CH", "LC"], ["CHF"])

    Also, the mapping can depending on the time period.

    Examples:
        For example, for periods before January 1999, we may want to map the
        local currency for Germany to `DEM` and afterwards to `EUR`. This can
        be expressed with the following value maps:

        >>> from datetime import datetime
        >>> t1 = datetime(1998, 12, 31, 23, 59, 59)
        >>> t2 = datetime(1999, 1, 1)
        >>> MultiValueMap(["DE", "LC"], ["EUR"], valid_to: t1)
        >>> MultiValueMap(["DE", "LC"], ["EUR"], valid_from: t2)
        >>> MultiValueMap(["CH", "LC"], ["CHF"])

    Values in the source may represent regular expressions with capture
    groups.

    Attributes:
        source: One or more source values
        target: One or more target values
        valid_from: Start of business validity for the mapping
        valid_to: End of business validity for the mapping
    """

    source: Sequence[Union[str, Pattern[str]]]
    target: Sequence[str]
    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None


class ValueMap(Struct, frozen=True, omit_defaults=True):
    """Maps the values of two components together.

    Examples:
        For example, let's assume that we want to map the ISO 2-letter country
        code with the ISO 3-letter country code. This can be expressed with
        the following value maps:

            >>> ValueMap("BE", "BEL")

    Values in the source may represent regular expressions with capture
    groups.

    Also, a map may have business validity associated with it.

    Attributes:
        source: The source value
        target: The target value
        valid_from: Start of business validity for the mapping
        valid_to: End of business validity for the mapping
    """

    source: Union[str, Pattern[str]]
    target: str
    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None


class MultiRepresentationMap(
    MaintainableArtefact, frozen=True, omit_defaults=True
):
    """Maps one or more source codelists to one or more target codelists.

    A representation map is iterable, i.e. it is possible to iterate over
    the various mappings using a `for` loop.

    Attributes:
        id: The identifier for the representation map.
        name: The representation map's name.
        agency: The maintainer of the representation map.
        source: The URN(s) of the source codelist(s) / valuelist(s),
            or data type(s).
        target: The URN(s) of the target codelist(s) / valuelist(s),
            or data type(s).
        maps: The various mappings in the representation map.
        description: Additional descriptive information about the
            representation map.
        version: The version of the representation map.
    """

    source: Sequence[Union[str, DataType]] = []
    target: Sequence[Union[str, DataType]] = []
    maps: Sequence[MultiValueMap] = []

    def __iter__(
        self,
    ) -> Iterator[MultiValueMap]:
        """Return an iterator over the different maps."""
        yield from self.maps

    def __len__(self) -> int:
        """Return the number of maps in the representation map."""
        return len(self.maps)


class MultiComponentMap(Struct, frozen=True, omit_defaults=True):
    """Maps one or more source components to one or more target components.

    Examples:
        For example, let's assume that we want to map the the code for local
        currency (say `LC`) to an ISO 3-letter currency code, depending on the
        country. So, if the country is `DE` (Germany), then the currency in
        the target should be `EUR` but, if the country is `CH`, then it should
        be `CHF`. This can be expressed as follows:

            >>> de = MultiValueMap(["DE", "LC"], ["EUR"])
            >>> ch = MultiValueMap(["CH", "LC"], ["CHF"])
            >>> src = ["COUNTRY", "CURRENCY"]
            >>> tgt = ["CURRENCY"]
            >>> cm = MultiComponentMap(src, tgt, [de, ch])

    Attributes:
        source: The source component(s)
        target: The target component(s)
        values: The representation map, with the list of mapped values
            (one or more in the source and one or more in the target)
    """

    source: Sequence[str]
    target: Sequence[str]
    values: MultiRepresentationMap


class RepresentationMap(MaintainableArtefact, frozen=True, omit_defaults=True):
    """Maps one source codelist to a target codelist.

    A representation map is iterable, i.e. it is possible to iterate over
    the various mappings using a `for` loop.

    Attributes:
        id: The identifier for the representation map.
        name: The representation map's name.
        agency: The maintainer of the representation map.
        source: The URN of the source codelist / valuelist or a data type.
        target: The URN of the target codelist / valuelist or a data type.
        maps: The various mappings in the representation map.
        description: Additional descriptive information about the
            representation map.
        version: The version of the representation map.
    """

    source: Union[str, DataType, None] = None
    target: Union[str, DataType, None] = None
    maps: Sequence[ValueMap] = []

    def __iter__(
        self,
    ) -> Iterator[ValueMap]:
        """Return an iterator over the different maps."""
        yield from self.maps

    def __len__(self) -> int:
        """Return the number of maps in the representation map."""
        return len(self.maps)


class ComponentMap(Struct, frozen=True, omit_defaults=True):
    """Maps a source component to a target component.

    Examples:
        For example, let's assume that we want to map the country code from
        an ISO 2-letter country code to an ISO 3-letter country code. This
        can be expressed as follows:

        >>> ar = ValueMap("AR", "ARG")
        >>> uy = ValueMap("UY", "URY")
        >>> cm = ComponentMap("COUNTRY", "COUNTRY", [ar, uy])

    Attributes:
        source: The source component
        target: The target component
        values: The representation map, with the list of mapped values
            (one in the source and one in the target)
    """

    source: str
    target: str
    values: RepresentationMap


class StructureMap(MaintainableArtefact, frozen=True, omit_defaults=True):
    """Maps a source structure to a target structure.

    The various mapping rules are classified by types.

    A structure map is an iterable, i.e. it is possible to iterate over
    the various mapping rules using a `for` loop.

    It is also possible to retrieve the mapping rule applying to a component
    by using the component id (e.g. `map["FREQ"]`).

    Attributes:
        id: The identifier for the structure map.
        name: The name of the structure map.
        agency: The maintainer of the structure map.
        source: The source structure.
        target: The target structure.
        maps: The various mapping rules in the structure map.
        description: Additional descriptive information about the structure
            map.
        version: The version of the structure map (e.g. 2.0.42).
    """

    source: str = ""
    target: str = ""
    maps: Sequence[
        Union[
            ComponentMap,
            DatePatternMap,
            FixedValueMap,
            ImplicitComponentMap,
            MultiComponentMap,
        ]
    ] = []

    @property
    def component_maps(self) -> Sequence[ComponentMap]:
        """Maps between one source and one target component."""
        return list(
            filter(
                lambda i: isinstance(  # type: ignore[arg-type]
                    i,
                    ComponentMap,
                ),
                self.maps,
            )
        )

    @property
    def date_pattern_maps(self) -> Sequence[DatePatternMap]:
        """Maps based on date patterns."""
        return list(
            filter(
                lambda i: isinstance(  # type: ignore[arg-type]
                    i,
                    DatePatternMap,
                ),
                self.maps,
            )
        )

    @property
    def fixed_value_maps(self) -> Sequence[FixedValueMap]:
        """Maps with a fixed value."""
        return list(
            filter(
                lambda i: isinstance(  # type: ignore[arg-type]
                    i,
                    FixedValueMap,
                ),
                self.maps,
            )
        )

    @property
    def implicit_component_maps(self) -> Sequence[ImplicitComponentMap]:
        """Maps where the source value is copied to the target."""
        return list(
            filter(
                lambda i: isinstance(  # type: ignore[arg-type]
                    i,
                    ImplicitComponentMap,
                ),
                self.maps,
            )
        )

    @property
    def multi_component_maps(self) -> Sequence[MultiComponentMap]:
        """Maps between one or more source & one or more target components."""
        return list(
            filter(
                lambda i: isinstance(  # type: ignore[arg-type]
                    i,
                    MultiComponentMap,
                ),
                self.maps,
            )
        )

    def __iter__(
        self,
    ) -> Iterator[
        Union[
            ComponentMap,
            DatePatternMap,
            FixedValueMap,
            ImplicitComponentMap,
            MultiComponentMap,
        ]
    ]:
        """Return an iterator over the different mapping rules."""
        yield from self.maps

    def __len__(self) -> int:
        """Return the number of mapping rules in the structure map."""
        return len(self.maps)

    def __getitem__(self, id_: str) -> Optional[
        Sequence[
            Union[
                ComponentMap,
                DatePatternMap,
                FixedValueMap,
                ImplicitComponentMap,
                MultiComponentMap,
            ]
        ]
    ]:
        """Return the mapping rules for the supplied component."""
        out = []
        for m in self.maps:
            if (
                hasattr(m, "source") and (m.source == id_ or id_ in m.source)
            ) or (isinstance(m, FixedValueMap) and m.target == id_):
                out.append(m)
        if len(out) == 0:
            return None
        else:
            return out



"""Schemas for Fusion-JSON messages."""

from pysdmx.io.json.fusion.messages.category import FusionCategorySchemeMessage
from pysdmx.io.json.fusion.messages.code import (
    FusionCodelistMessage,
    FusionHierarchyAssociationMessage,
    FusionHierarchyMessage,
)
from pysdmx.io.json.fusion.messages.concept import FusionConceptSchemeMessage
from pysdmx.io.json.fusion.messages.dataflow import FusionDataflowMessage
from pysdmx.io.json.fusion.messages.map import (
    FusionMappingMessage,
    FusionRepresentationMapMessage,
)
from pysdmx.io.json.fusion.messages.org import (
    FusionAgencyMessage,
    FusionProviderMessage,
)
from pysdmx.io.json.fusion.messages.report import FusionMetadataMessage
from pysdmx.io.json.fusion.messages.schema import FusionSchemaMessage

__all__ = [
    "FusionCategorySchemeMessage",
    "FusionCodelistMessage",
    "FusionHierarchyAssociationMessage",
    "FusionHierarchyMessage",
    "FusionConceptSchemeMessage",
    "FusionDataflowMessage",
    "FusionMappingMessage",
    "FusionRepresentationMapMessage",
    "FusionAgencyMessage",
    "FusionProviderMessage",
    "FusionMetadataMessage",
    "FusionSchemaMessage",
]
